@@ -1,6 +1,7 @@
 /* Fault handler information.  Woe32 version.
-   Copyright (C) 1993-1999, 2002-2003, 2007-2008  Bruno Haible <bruno@clisp.org>
+   Copyright (C) 1993-1999, 2002-2003, 2007-2009, 2011, 2016  Bruno Haible <bruno@clisp.org>
    Copyright (C) 2003  Paolo Bonzini <bonzini@gnu.org>
+   Copyright (C) 2009  Eric Blake <ebb9@byu.net>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,11 +29,45 @@
  * extern DWORD GetLastError (void);
  */
 
+/* On Cygwin, the "system calls" are actually library calls.  If the user
+   asks for them to return with errno = EFAULT, we do this by combining
+   the Win32 and the Unix approaches:
+     1. Install a Win32 exception filter and use it to peek at the fault
+        context, but without actually handling the fault.
+     2. Install a Unix signal handler.
+   If a fault occurs outside a "system call", our exception filter will
+   take a note and defer, Cygwin's exception filter will invoke our Unix
+   signal handler.
+   If a fault occurs inside a "system call", our exception filter will
+   take a note and defer, Cygwin's exception filter will handle it, and
+   our signal handler will not be called.
+   For stack overflow, we use the pure Win32 approach, since Cygwin does
+   not have sigaltstack().  */
+#if defined __CYGWIN__ && ENABLE_EFAULT
+
+# define MIXING_UNIX_SIGSEGV_AND_WIN32_STACKOVERFLOW_HANDLING 1
+
+# if OLD_CYGWIN_WORKAROUND
+/* Last seen fault address.  */
+static void *last_seen_fault_address;
+# endif
+
+/* Import the relevant part of the Unix approach.  */
+# undef HAVE_STACK_OVERFLOW_RECOVERY
+# define HAVE_STACK_OVERFLOW_RECOVERY 0
+# define sigsegv_install_handler sigsegv_install_handler_unix
+# include "handler-unix.c"
+# undef sigsegv_install_handler
+
+#else
+
 /* User's SIGSEGV handler.  */
 static sigsegv_handler_t user_handler = (sigsegv_handler_t) NULL;
 
+#endif
+
 /* Stack overflow handling is tricky:
-   First, we must catch a STATUS_STACK_OVERFLOW exception. This is signalled
+   First, we must catch a EXCEPTION_STACK_OVERFLOW exception. This is signalled
    when the guard page at the end of the stack has been touched. The operating
    system remaps the page with protection PAGE_READWRITE and only then calls
    our exception handler. Actually, it's even more complicated: The stack has
@@ -64,22 +99,22 @@ static sigsegv_handler_t user_handler = (sigsegv_handler_t) NULL;
 
 static stackoverflow_handler_t stk_user_handler =
   (stackoverflow_handler_t) NULL;
-static unsigned long stk_extra_stack;
-static unsigned long stk_extra_stack_size;
+static uintptr_t stk_extra_stack;
+static size_t stk_extra_stack_size;
 
 static void
-stack_overflow_handler (unsigned long faulting_page_address, stackoverflow_context_t context)
+stack_overflow_handler (uintptr_t faulting_page_address, stackoverflow_context_t context)
 {
   MEMORY_BASIC_INFORMATION info;
   DWORD oldprot;
-  unsigned long base;
-  unsigned long address;
+  uintptr_t base;
+  uintptr_t address;
 
   /* First get stack's base address.  */
   if (VirtualQuery ((void*) faulting_page_address, &info, sizeof (info))
       != sizeof (info))
     goto failed;
-  base = (unsigned long) info.AllocationBase;
+  base = (uintptr_t) info.AllocationBase;
 
   /* Now search for the first existing page.  */
   address = base;
@@ -87,16 +122,16 @@ stack_overflow_handler (unsigned long faulting_page_address, stackoverflow_conte
     {
       if (VirtualQuery ((void*) address, &info, sizeof (info)) != sizeof (info))
         goto failed;
-      if (address != (unsigned long) info.BaseAddress)
+      if (address != (uintptr_t) info.BaseAddress)
        goto failed;
       if (info.State != MEM_FREE)
         {
-          if ((unsigned long) info.AllocationBase != base)
+          if ((uintptr_t) info.AllocationBase != base)
             goto failed;
           if (info.State == MEM_COMMIT)
             break;
         }
-      address = (unsigned long) info.BaseAddress + info.RegionSize;
+      address = (uintptr_t) info.BaseAddress + info.RegionSize;
     }
 
   /* Now add the PAGE_GUARD bit to the first existing page.  */
@@ -121,12 +156,17 @@ static LONG WINAPI
 main_exception_filter (EXCEPTION_POINTERS *ExceptionInfo)
 {
   if ((stk_user_handler
-       && ExceptionInfo->ExceptionRecord->ExceptionCode == STATUS_STACK_OVERFLOW
+       && ExceptionInfo->ExceptionRecord->ExceptionCode == EXCEPTION_STACK_OVERFLOW
       )
+#if !MIXING_UNIX_SIGSEGV_AND_WIN32_STACKOVERFLOW_HANDLING || OLD_CYGWIN_WORKAROUND
       ||
-      (user_handler != (sigsegv_handler_t)NULL
-       && ExceptionInfo->ExceptionRecord->ExceptionCode == EXCEPTION_ACCESS_VIOLATION
-     ))
+      (ExceptionInfo->ExceptionRecord->ExceptionCode == EXCEPTION_ACCESS_VIOLATION
+#if !MIXING_UNIX_SIGSEGV_AND_WIN32_STACKOVERFLOW_HANDLING
+       && user_handler != (sigsegv_handler_t)NULL
+#endif
+      )
+#endif
+     )
     {
 #if 0 /* for debugging only */
       printf ("Exception!\n");
@@ -158,7 +198,7 @@ main_exception_filter (EXCEPTION_POINTERS *ExceptionInfo)
       if (ExceptionInfo->ExceptionRecord->NumberParameters == 2)
         {
           if (stk_user_handler
-              && ExceptionInfo->ExceptionRecord->ExceptionCode == STATUS_STACK_OVERFLOW)
+              && ExceptionInfo->ExceptionRecord->ExceptionCode == EXCEPTION_STACK_OVERFLOW)
             {
               char *address = (char *) ExceptionInfo->ExceptionRecord->ExceptionInformation[1];
               /* Restart the program, giving it a sane value for %esp.
@@ -167,8 +207,8 @@ main_exception_filter (EXCEPTION_POINTERS *ExceptionInfo)
                  to be allocated in the guard page, where it will be
                  inaccessible as soon as we restore the PAGE_GUARD bit!) to
                  this new stack.  */
-              unsigned long faulting_page_address = (unsigned long)address & -0x1000;
-              unsigned long new_safe_esp = ((stk_extra_stack + stk_extra_stack_size) & -16);
+              uintptr_t faulting_page_address = (uintptr_t)address & -0x1000;
+              uintptr_t new_safe_esp = ((stk_extra_stack + stk_extra_stack_size) & -16);
               CONTEXT *orig_context = ExceptionInfo->ContextRecord;
               CONTEXT *safe_context = (CONTEXT *) (new_safe_esp -= sizeof (CONTEXT)); /* make room */
               memcpy (safe_context, orig_context, sizeof (CONTEXT));
@@ -177,21 +217,35 @@ main_exception_filter (EXCEPTION_POINTERS *ExceptionInfo)
               new_safe_esp -= 4; /* make room for (unused) return address slot */
               ExceptionInfo->ContextRecord->Esp = new_safe_esp;
               /* Call stack_overflow_handler(faulting_page_address,safe_context).  */
-              ExceptionInfo->ContextRecord->Eip = (unsigned long)&stack_overflow_handler;
-              *(unsigned long *)(new_safe_esp + 4) = faulting_page_address;
-              *(unsigned long *)(new_safe_esp + 8) = (unsigned long) safe_context;
+              ExceptionInfo->ContextRecord->Eip = (uintptr_t)&stack_overflow_handler;
+              *(uintptr_t *)(new_safe_esp + 4) = faulting_page_address;
+              *(uintptr_t *)(new_safe_esp + 8) = (uintptr_t) safe_context;
               return EXCEPTION_CONTINUE_EXECUTION;
             }
-          if (user_handler != (sigsegv_handler_t) NULL
-              && ExceptionInfo->ExceptionRecord->ExceptionCode == EXCEPTION_ACCESS_VIOLATION)
+#if !MIXING_UNIX_SIGSEGV_AND_WIN32_STACKOVERFLOW_HANDLING || OLD_CYGWIN_WORKAROUND
+          if (ExceptionInfo->ExceptionRecord->ExceptionCode == EXCEPTION_ACCESS_VIOLATION)
             {
-              /* ExceptionInfo->ExceptionRecord->ExceptionInformation[0] is 1
-                 if it's a write access, 0 if it's a read access. But we don't
-                 need this info because we don't have it on Unix either.  */
-              void *address = (void *) ExceptionInfo->ExceptionRecord->ExceptionInformation[1];
-              if ((*user_handler) (address, 1))
-                return EXCEPTION_CONTINUE_EXECUTION;
+#if MIXING_UNIX_SIGSEGV_AND_WIN32_STACKOVERFLOW_HANDLING
+              /* Store the fault address.  Then pass control to Cygwin's
+                 exception filter, which will decide whether to recognize
+                 a fault inside a "system call" (and return errno = EFAULT)
+                 or to pass it on to the program (by invoking the Unix signal
+                 handler).  */
+              last_seen_fault_address = (void *) ExceptionInfo->ExceptionRecord->ExceptionInformation[1];
+#else
+              if (user_handler != (sigsegv_handler_t) NULL)
+                {
+                  /* ExceptionInfo->ExceptionRecord->ExceptionInformation[0] is
+                     1 if it's a write access, 0 if it's a read access. But we
+                     don't need this info because we don't have it on Unix
+                     either.  */
+                  void *address = (void *) ExceptionInfo->ExceptionRecord->ExceptionInformation[1];
+                  if ((*user_handler) (address, 1))
+                    return EXCEPTION_CONTINUE_EXECUTION;
+                }
+#endif
             }
+#endif
         }
     }
   return EXCEPTION_CONTINUE_SEARCH;
@@ -223,7 +277,11 @@ __asm__ (".equ __except_list,0");
 extern exception_list *_except_list __asm__ ("%fs:__except_list");
 
 /* For debugging.  _except_list is not otherwise accessible from gdb.  */
-static exception_list *
+static
+#if __GNUC__ >= 3
+__attribute__ ((__unused__)) /* avoid GCC warning */
+#endif
+exception_list *
 debug_get_except_list ()
 {
   return _except_list;
@@ -277,6 +335,17 @@ install_main_exception_filter ()
     }
 }
 
+#if MIXING_UNIX_SIGSEGV_AND_WIN32_STACKOVERFLOW_HANDLING
+
+int
+sigsegv_install_handler (sigsegv_handler_t handler)
+{
+  install_main_exception_filter ();
+  return sigsegv_install_handler_unix (handler);
+}
+
+#else
+
 int
 sigsegv_install_handler (sigsegv_handler_t handler)
 {
@@ -299,12 +368,14 @@ sigsegv_leave_handler (void (*continuation) (void*, void*, void*),
   return 1;
 }
 
+#endif /* !MIXING_UNIX_SIGSEGV_AND_WIN32_STACKOVERFLOW_HANDLING */
+
 int
 stackoverflow_install_handler (stackoverflow_handler_t handler,
-                               void *extra_stack, unsigned long extra_stack_size)
+                               void *extra_stack, size_t extra_stack_size)
 {
   stk_user_handler = handler;
-  stk_extra_stack = (unsigned long) extra_stack;
+  stk_extra_stack = (uintptr_t) extra_stack;
   stk_extra_stack_size = extra_stack_size;
   install_main_exception_filter ();
   return 0;
