@@ -1,5 +1,5 @@
 /* Test that the handler is called, with the right fault address.
-   Copyright (C) 2002-2005  Bruno Haible <bruno@clisp.org>
+   Copyright (C) 2002-2006, 2008, 2011, 2016  Bruno Haible <bruno@clisp.org>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,9 +13,14 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
+
+#ifndef _MSC_VER
+# include <config.h>
+#endif
 
 #include "sigsegv.h"
+#include <stdint.h>
 #include <stdio.h>
 
 #if HAVE_SIGSEGV_RECOVERY
@@ -23,7 +28,7 @@
 #include "mmaputil.h"
 #include <stdlib.h>
 
-unsigned long page;
+uintptr_t page;
 
 volatile int handler_called = 0;
 
@@ -33,7 +38,8 @@ handler (void *fault_address, int serious)
   handler_called++;
   if (handler_called > 10)
     abort ();
-  if (fault_address != (void *)(page + 0x678))
+  if (fault_address
+      != (void *)((page + 0x678) & ~(SIGSEGV_FAULT_ADDRESS_ALIGNMENT - 1)))
     abort ();
   if (mprotect ((void *) page, 0x4000, PROT_READ_WRITE) == 0)
     return 1;
@@ -41,14 +47,15 @@ handler (void *fault_address, int serious)
 }
 
 void
-crasher (unsigned long p)
+crasher (uintptr_t p)
 {
-  *(int *) (p + 0x678) = 42;
+  *(volatile int *) (p + 0x678) = 42;
 }
 
 int
 main ()
 {
+  int prot_unwritable;
   void *p;
 
   /* Preparations.  */
@@ -56,6 +63,14 @@ main ()
   zero_fd = open ("/dev/zero", O_RDONLY, 0644);
 #endif
 
+#if defined __linux__ && defined __sparc__
+  /* On Linux 2.6.26/SPARC64, PROT_READ has the same effect as
+     PROT_READ | PROT_WRITE.  */
+  prot_unwritable = PROT_NONE;
+#else
+  prot_unwritable = PROT_READ;
+#endif
+
   /* Setup some mmaped memory.  */
   p = mmap_zeromap ((void *) 0x12340000, 0x4000);
   if (p == (void *)(-1))
@@ -63,10 +78,10 @@ main ()
       fprintf (stderr, "mmap_zeromap failed.\n");
       exit (2);
     }
-  page = (unsigned long) p;
+  page = (uintptr_t) p;
 
   /* Make it read-only.  */
-  if (mprotect ((void *) page, 0x4000, PROT_READ) < 0)
+  if (mprotect ((void *) page, 0x4000, prot_unwritable) < 0)
     {
       fprintf (stderr, "mprotect failed.\n");
       exit (2);
@@ -74,7 +89,7 @@ main ()
   /* Test whether it's possible to make it read-write after it was read-only.
      This is not possible on Cygwin.  */
   if (mprotect ((void *) page, 0x4000, PROT_READ_WRITE) < 0
-      || mprotect ((void *) page, 0x4000, PROT_READ) < 0)
+      || mprotect ((void *) page, 0x4000, prot_unwritable) < 0)
     {
       fprintf (stderr, "mprotect failed.\n");
       exit (2);
