@@ -1,5 +1,6 @@
 /* Test the stack overflow handler.
-   Copyright (C) 2002-2006, 2008  Bruno Haible <bruno@clisp.org>
+   Copyright (C) 2002-2006, 2008, 2010  Bruno Haible <bruno@clisp.org>
+   Copyright (C) 2010 Eric Blake <eblake@redhat.com>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -41,9 +42,7 @@
 # include <sys/time.h>
 # include <sys/resource.h>
 #endif
-#ifndef SIGSTKSZ
-# define SIGSTKSZ 16384
-#endif
+#include "altstack.h"
 
 jmp_buf mainloop;
 sigset_t mainsigset;
@@ -88,10 +87,6 @@ recurse (volatile int n)
   return *recurse_1 (n, &n);
 }
 
-/* glibc says: Users should use SIGSTKSZ as the size of user-supplied
-   buffers.  */
-char mystack[SIGSTKSZ];
-
 int
 main ()
 {
@@ -106,13 +101,16 @@ main ()
   setrlimit (RLIMIT_STACK, &rl);
 #endif
 
+  /* Prepare the storage for the alternate stack.  */
+  prepare_alternate_stack ();
+
   /* Install the stack overflow handler.  */
   if (stackoverflow_install_handler (&stackoverflow_handler,
-                                     mystack, sizeof (mystack))
+                                     mystack, SIGSTKSZ)
       < 0)
     exit (2);
   stack_lower_bound = mystack;
-  stack_upper_bound = mystack + sizeof (mystack) - 1;
+  stack_upper_bound = mystack + SIGSTKSZ - 1;
 
   /* Save the current signal mask.  */
   sigemptyset (&emptyset);
@@ -133,6 +131,9 @@ main ()
       abort ();
     }
 
+  /* Validate that the alternate stack did not overflow.  */
+  check_alternate_stack_no_overflow ();
+
   printf ("Test passed.\n");
   exit (0);
 }
